@@ -17,8 +17,15 @@ from hyperopt.fmin import generate_trials_to_calculate
 import logging, pickle
 import json
 import datetime
+from poap.controller import SerialController, EvalRecord
+from pySOT.optimization_problems import OptimizationProblem
+from pySOT.experimental_design import SymmetricLatinHypercube
+from pySOT.strategy import SRBFStrategy, DYCORSStrategy
+from pySOT.surrogate import CubicKernel, LinearTail, RBFInterpolant
+from lipo import GlobalOptimizer
 from matplotlib import pyplot as plt
 import seaborn as sns; sns.set()
+from collections import OrderedDict
 logger = logging.getLogger(__name__)
 
 # import experiments
@@ -141,13 +148,103 @@ def ku_pdf(X, a, b):
     return pdf
 
 
+def get_sample_and_weights_using_oracle(sample_size, X, y, uncertainties, dp_alpha, prior_for_a_beta_A,
+                      prior_for_a_beta_B, prior_for_b_beta_A, prior_for_b_beta_B, scale_a=10, scale_b=10,
+                       pct_from_original=0.0, num_trials=1000):
+    """
+        Given the training data, generate samples (w/o replacment) based on the oracle. The number of points returned
+        in the sample may be less than what is requested since the Beta might be concentrated in regions where we
+        have no points; quite possible in the early parts of the search for the optimal distribution.
+
+        :param sample_size: number of points we want to sample
+        :param X: 2D-array of data to sample from
+        :param y: labels corresponding to X
+        :param dp_alpha: concentration parameter of Dirichlet Process
+        :param prior_for_a_beta_A: Beta shape parameter A for the prior to use to sample a
+        :param prior_for_a_beta_B: Beta shape parameter B for the prior to use to sample a
+        :param prior_for_b_beta_A: Beta shape parameter A for the prior to use to sample b
+        :param prior_for_b_beta_B: Beta shape parameter B for the prior to use to sample b
+        :param scale_a: scaling parameter for Beta, see ``scipy.stats.beta``
+        :param scale_b: scaling parameter for Beta, see ``scipy.stats.beta``
+        :param pct_from_original: percentage of points to be sampled from the original distribution. This is a stratified
+            sample
+        :param num_trials: # trials over which instance weights are to be averaged
+        :return: sampled X, y, and vector of corresponding instance weights.
+        Number of points in the sample might be less than sample_size
+        """
+    sample_size_from_oracle, sample_size_from_orig = sample_size, 0
+    pct_from_original = min([1.0, pct_from_original])
+    pct_from_oracle = 1.0 - pct_from_original
+    if 0 < pct_from_original:
+        sample_size_from_orig = int(pct_from_original * sample_size)
+        sample_size_from_oracle = sample_size - sample_size_from_orig
+    float_max = np.finfo(np.float64).max
+    sample_X, sample_y, trial_weights = [], [], np.zeros((num_trials, len(uncertainties)))
+    if pct_from_oracle > 0:
+        logging.info(f"Starting to sample from oracle, over {num_trials} trials.")
+        start_time = datetime.datetime.now()
+        for tr_idx in range(num_trials):
+            # determine the IBMM
+            clusters = dp.generic_blackwell_macqueen_sampler(dp_alpha, sample_size_from_oracle)
+            betas_for_sampling = dp.sample_beta(prior_for_a_beta_A, prior_for_a_beta_B, prior_for_b_beta_A,
+                                                prior_for_b_beta_B, len(clusters), scale_a, scale_b)
+            for cluster_size, indv_beta in zip(clusters, betas_for_sampling):
+                if cluster_size == 0:
+                    continue
+                frozen_pdf = beta(indv_beta[0], indv_beta[1])
+                sample_probs = frozen_pdf.pdf(uncertainties)
+                # sample_probs = np.clip(sample_probs, a_min=np.min(sample_probs), a_max=float_max)
+                trial_weights[tr_idx, :] = sample_probs
+        end_time = datetime.datetime.now()
+        duration = (end_time - start_time).total_seconds()
+        logging.info(
+            f"Finished obtaining weights from oracle, over {num_trials} trials, {1.0 * duration / num_trials} sec per trial.")
+
+        # we need to average weights across trials - but need to be mindful about the fact that we might see np.inf
+        # values here's what we do: replace infs with the max float value *after* taking the mean. Same with the total
+        # - note: the replacement in the pdf weights should be done after computing the total, otherwise np.sum
+        # (used to compute total) might again bump up values to np.inf
+        instance_weights = np.mean(trial_weights, axis=0)
+        total_weight = np.sum(instance_weights)
+        instance_weights = np.clip(instance_weights, a_min=np.min(instance_weights), a_max=float_max)
+        if np.isinf(total_weight):
+            total_weight = float_max
+        normalized_oracle_weights = instance_weights/total_weight
+        if np.any(np.isinf(normalized_oracle_weights)):
+            logging.debug(f"Found inf in normalized instance weight array.")
+        if np.any(np.isnan(normalized_oracle_weights)):
+            logging.debug(f"Found nan in normalized instance weight array.")
+    else:
+        normalized_oracle_weights =  np.zeros(len(uncertainties))
+
+    # get weights from original, this is simply 1/N
+    normalized_orig_weights = np.ones(len(uncertainties))/len(uncertainties)
+
+    # get the final weights
+    normalized_weights = pct_from_oracle * normalized_oracle_weights + pct_from_original * normalized_orig_weights
+    if np.any(np.isinf(normalized_weights)):
+        logging.debug(f"Found inf in normalized weight array.")
+    if np.any(np.isnan(normalized_weights)):
+        logging.debug(f"Found nan in normalized weight array.")
+
+    # get the desired sample size after ignoring zero weights (zero weights are impossible if pct_from_original > 0 )
+    top_idxs = np.argsort(normalized_weights)[-sample_size:]
+    top_idxs = [idx for idx in top_idxs if normalized_weights[idx] > 0]
+    gap = sample_size - len(top_idxs)
+    if gap > 0:
+        logging.debug("Requested total size: %d, sampled size=%d, skipped=%d" % (sample_size,
+                                                                       len(top_idxs), gap))
+    sample_X, sample_y, sample_weights = X[top_idxs, :], y[top_idxs], normalized_weights[top_idxs]
+    return sample_X, sample_y, sample_weights
+
+
 def sample_using_oracle(sample_size, X, y, uncertainties, dp_alpha, prior_for_a_beta_A,
                       prior_for_a_beta_B, prior_for_b_beta_A, prior_for_b_beta_B, scale_a=10, scale_b=10,
-                       pct_from_original=0.0):
+                       pct_from_original=0.0, use_weights=False, num_weight_trials=10):
     """
-    Given the training data, generate samples based on the oracle. The number of points returned in the sample may be
-    less than what is requested since the Beta might be concentrated in regions where we have no points; quite possible
-    in the early parts of the search for the optimal distribution.
+    Given the training data, generate samples (with replacement) based on the oracle. The number of points returned in
+    the sample may be less than what is requested since the Beta might be concentrated in regions where we have no
+    points; quite possible in the early parts of the search for the optimal distribution.
 
     :param sample_size: number of points we want to sample
     :param X: 2D-array of data to sample from
@@ -163,6 +260,15 @@ def sample_using_oracle(sample_size, X, y, uncertainties, dp_alpha, prior_for_a_
         sample
     :return: sampled X and y. Number of points in the sample might be less than sample_size
     """
+    if use_weights:
+        sample_X, sample_y, sample_weights = get_sample_and_weights_using_oracle(sample_size, X, y, uncertainties,
+                                                                                 dp_alpha, prior_for_a_beta_A,
+                                                                                 prior_for_a_beta_B, prior_for_b_beta_A,
+                                                                                 prior_for_b_beta_B, scale_a,
+                                                                                 scale_b,
+                                                                                 pct_from_original,
+                                                                                 num_trials=num_weight_trials)
+        return sample_X, sample_y, sample_weights
     sample_size_from_oracle, sample_size_from_orig = sample_size, 0
     pct_from_original = min([1.0, pct_from_original])
     if 0 < pct_from_original:
@@ -248,8 +354,8 @@ def adjust_sample_from_dist(sample_size, pct_from_original, min_sample_size_per_
     return pct_from_original
 
 
-def create_hyperopt_search_space(max_components, min_sample_size, min_sample_size_as_pct, max_sample_size_as_pct,
-                                 alpha_multiplier=2.0, extended_space=True):
+def create_search_space(max_components, min_sample_size, min_sample_size_as_pct, max_sample_size_as_pct,
+                        alpha_multiplier=2.0, extended_space=True, optimizer_type='hyperopt'):
     """
     Creates a list of search spaces using hyperopt primitives.
 
@@ -263,6 +369,7 @@ def create_hyperopt_search_space(max_components, min_sample_size, min_sample_siz
         sample size
     :param alpha_multiplier: see notes within function
     :param extended_space: if True the shape prior param space is considerable extended
+    :param optimizer_type: based on the optimizer backend a different representation of the search space is returned.
     :return: list of hyperopt search space primitives
     """
     # H_n = dp.harmonic_num(min_sample_size)  # H_n is the nth harmonic number
@@ -275,27 +382,55 @@ def create_hyperopt_search_space(max_components, min_sample_size, min_sample_siz
     # max_DP_alpha = alpha_multiplier * max_components / H_n
     max_DP_alpha = (max_components - 2.0) /5.0
     space = [hp.uniform('dp_alpha', 0.1, max_DP_alpha)]
+    only_unif = OrderedDict({'dp_alpha': (0.1, max_DP_alpha)})  # only uniform search spaces
+    log_transforms = []  # these uniform variables need to be transformed to log-uniform
 
     # the optimizer decides the size of the sample to be used to train the small model; this is expressed as a fraction
     # of the size of the training data
     space.append(hp.uniform('sample_size_as_pct', min_sample_size_as_pct, max_sample_size_as_pct))
+    only_unif['sample_size_as_pct'] = (min_sample_size_as_pct, max_sample_size_as_pct)
 
     # for a given sample size, the optimizer is free to pick some proportion from the original dataset too
     space.append(hp.uniform('pct_from_original', 0.0, 1.0))
+    only_unif['pct_from_original'] = (0.0, 1.0)
 
     # the priors for components
+    # NOTE: the log transformation needs to be done in the optimizer in the case of uniform search space!
     if extended_space:
         space.append(hp.loguniform('prior_for_a_beta_A', -10, 10))
+        only_unif['prior_for_a_beta_A'] = (-10, 10)
+        log_transforms.append('prior_for_a_beta_A')
+
         space.append(hp.loguniform('prior_for_a_beta_B', -10, 10))
+        only_unif['prior_for_a_beta_B'] = (-10, 10)
+        log_transforms.append('prior_for_a_beta_B')
+
         space.append(hp.loguniform('prior_for_b_beta_A', -10, 10))
+        only_unif['prior_for_b_beta_A'] = (-10, 10)
+        log_transforms.append('prior_for_b_beta_A')
+
         space.append(hp.loguniform('prior_for_b_beta_B', -10, 10))
+        only_unif['prior_for_b_beta_B'] = (-10, 10)
+        log_transforms.append('prior_for_b_beta_B')
     else:
         space.append(hp.uniform('prior_for_a_beta_A', 0.1, 100))
-        space.append(hp.uniform('prior_for_a_beta_B', 0.1, 100))
-        space.append(hp.uniform('prior_for_b_beta_A', 0.1, 100))
-        space.append(hp.uniform('prior_for_b_beta_B', 0.1, 100))
+        only_unif['prior_for_a_beta_A'] = (0.1, 100)
 
-    return space
+        space.append(hp.uniform('prior_for_a_beta_B', 0.1, 100))
+        only_unif['prior_for_a_beta_B'] = (0.1, 100)
+
+        space.append(hp.uniform('prior_for_b_beta_A', 0.1, 100))
+        only_unif['prior_for_b_beta_A'] = (0.1, 100)
+
+        space.append(hp.uniform('prior_for_b_beta_B', 0.1, 100))
+        only_unif['prior_for_b_beta_B'] = (0.1, 100)
+
+    if optimizer_type == 'hyperopt':
+        return space
+    elif optimizer_type in {'pysot', 'lipo'}:
+        return only_unif, log_transforms
+    else:
+        return None
 
 
 def determine_split_type(splits):
@@ -330,7 +465,7 @@ def determine_split_type(splits):
 def run_experiment(dataset, model_obj, oracle,
                    evals=100, min_sample_size=100, min_sample_size_per_dist=100, max_sample_size=None,
                    max_components=30,
-                   sampling_trials=1,
+                   sampling_trials=1, use_weights=False, num_weight_trials=10,
                    splits=None,
                    dataset_name=None,
                    create_persistence_obj=False, use_persistence_obj=None,
@@ -339,7 +474,9 @@ def run_experiment(dataset, model_obj, oracle,
                    additional_init_vals=None,
                    scale=10000,
                    compute_all_baselines=False,
-                   save_optimal_sample=False):
+                   save_optimal_sample=False,
+                   use_extended_search_space=True,
+                   optimizer='hyperopt', optimizer_params=None):
     """
     Runs the optimization for given dataset, splits, oracle etc. This is the core algorithm.
 
@@ -363,7 +500,12 @@ def run_experiment(dataset, model_obj, oracle,
     :param max_sample_size: maximum number of points that can be sampled, if this is None, the upper limit is the size
         of the training split
     :param max_components: max number of components of the mixture model (over uncertainties)
-    :param sampling_trials: for a particular setting of the DP how many times to sample and evaluate? this is for rigor
+    :param sampling_trials: for a particular setting of the DP how many times to sample, fit and evaluate?
+        This is for rigor - averaging over samples provides a stable score for the optimizer to use.
+    :param use_weights: you can avoid sampling and fitting (which can be expensive) by using instance weights instead,
+        IF your classifier supports learning using instance weights. If this is set to True, sampling_trials MUST be
+        set to 1
+    :param num_weight_trials: if use_weights is True, this is the number of times instance weights would be averaged
     :param splits: what fraction of the dataset to be used for various tests. This can take on a few possible types:
 
         *   None: the train-val-test split is the same as DEFAULT_SPLITS (see function body)
@@ -401,6 +543,11 @@ def run_experiment(dataset, model_obj, oracle,
         taking and are not useful for most of the practical use-cases.
     :param save_optimal_sample: whether the samples at the best validation score to be saved into a file. Note that
         since there are more than one sampling trials, all of them would be saved.
+    :param use_extended_search_space: solution is searched over a larger space
+    :param optimizer: which optimizer backend to use? This can be "hyperopt", "pysot" or "lipo".
+    :param optimizer_params: if the optimizer accepts params, they need to be passed in as a dict. There is no standard
+        format for this, and would be parsed in an optimizer specific way. Current supported:
+        * for optimizer=pysot, this parameter can be {"strategy": "srbf"} or {"strategy": "dycors"}
 
     :return: *test improvement at the best validation accuracy, best parameters, new_persistence_obj*.
 
@@ -410,6 +557,14 @@ def run_experiment(dataset, model_obj, oracle,
     """
     ERRORED_SAVE_PATH = "Error: couldn't save file."
     DEFAULT_SPLITS = {'train': 0.6, 'val': 0.2, 'test': 0.2}
+    SUPPORTED_OPTIMIZERS = {'hyperopt', 'pysot', 'lipo'}
+    OPTVAR_DIMS = 7
+    if optimizer not in SUPPORTED_OPTIMIZERS:
+        logging.error(f"Unsupported optimizer: {optimizer}, we only support {','.join(SUPPORTED_OPTIMIZERS)}")
+        return
+    if use_weights and sampling_trials > 1:
+        logging.warning(f"Since use_weights={use_weights}, sampling_trials is set to 1.")
+        sampling_trials = 1
     if splits is None:
         splits = DEFAULT_SPLITS
 
@@ -461,8 +616,11 @@ def run_experiment(dataset, model_obj, oracle,
 
     result_json = {"general_info": {"complexity_param": str(model_obj.complexity_param),
                                     "complexity_param_pickled_hex": pickle.dumps(model_obj.complexity_param).hex(),
-                                    "optimizer_iterations": evals,
-                   "model": model_obj.__class__.__name__,
+                                    "optimizer_iterations": evals, "optimizer": optimizer,
+                                    "optimizer_params": optimizer_params,
+                                    "sampling_style": f"weight averaging ({num_weight_trials})" if use_weights else
+                                                        f"sample and fit ({sampling_trials})",
+                                    "model": model_obj.__class__.__name__,
                                     'oracle': oracle_str}}
 
     X, y = None, None
@@ -617,6 +775,7 @@ def run_experiment(dataset, model_obj, oracle,
     best_samples_so_far = [None]
 
     def minimization_objective(params):
+        eval_start_time = datetime.datetime.now()
         dp_alpha, sample_size_as_pct, pct_from_original, prior_for_a_beta_A, prior_for_a_beta_B, \
         prior_for_b_beta_A, prior_for_b_beta_B = params
         sample_size = int(math.floor(sample_size_as_pct * np.shape(X_train)[0]))
@@ -628,19 +787,28 @@ def run_experiment(dataset, model_obj, oracle,
         optimizer_samples = []
         for i in range(sampling_trials):
             logging.info("Running sampling trial: %d of %d" % (i + 1, sampling_trials))
-            sample_X, sample_y = sample_using_oracle(sample_size, X_train, y_train, uncertainties,
+            temp_sample_info = sample_using_oracle(sample_size, X_train, y_train, uncertainties,
                                                    dp_alpha, prior_for_a_beta_A,
                                                    prior_for_a_beta_B, prior_for_b_beta_A,
                                                    prior_for_b_beta_B,
-                                                   pct_from_original=pct_from_original, scale_a=scale, scale_b=scale)
-            optimizer_samples.append((sample_X, sample_y))
+                                                   pct_from_original=pct_from_original, scale_a=scale, scale_b=scale,
+                                                     use_weights=use_weights, num_weight_trials=num_weight_trials)
+
+            sample_X, sample_y = temp_sample_info[:2]
+            if use_weights:
+                sample_weights = temp_sample_info[2]
+
+            optimizer_samples.append(temp_sample_info)
             num_unique_lables_in_sample = len(set(sample_y))
             if num_unique_lables_in_sample< 2:
                 logging.debug("Sample has only 1 label!")
 
             if len(sample_X) > 0 and num_unique_lables_in_sample > 1:
                 logging.info("Fitting model within optimizer iteration.")
-                temp_ret = model_obj.fit_model_within_iteration(sample_X, sample_y)
+                if use_weights:
+                    temp_ret = model_obj.fit_model_within_iteration(sample_X, sample_y, weights=sample_weights)
+                else:
+                    temp_ret = model_obj.fit_model_within_iteration(sample_X, sample_y)
                 if temp_ret is None:
                     logging.debug("Fit within iteration returned None.")
                     train_trial_score, val_trial_score, test_trial_score, current_iter_model, current_iter_params = \
@@ -676,11 +844,12 @@ def run_experiment(dataset, model_obj, oracle,
                                                                                          mean_test_score,
                                                                                         best_mean_val_score_so_far[0]))
         loss = 1 - mean_train_score
+        eval_end_time = datetime.datetime.now()
+        eval_duration_sec = (eval_end_time - eval_start_time).total_seconds()
         return {'loss': loss, 'status': STATUS_OK, 'mean_train_score': mean_train_score,
                 'mean_val_score': mean_val_score, 'mean_test_score': mean_test_score,
-                'avg_complexity': avg_complexity}
-
-    # run the optimizer
+                'avg_complexity': avg_complexity, 'params': list(params),
+                'start_time': str(eval_start_time), 'end_time': str(eval_end_time), 'duration_in_sec':eval_duration_sec}
 
     # initialize with one point - this represents the accuracy we'd get with the original dataset
     # we only need to ensure that sample_size_as_pct=1 and pct_from_original=1. The other values may be arbitrary
@@ -689,30 +858,147 @@ def run_experiment(dataset, model_obj, oracle,
     max_sample_size_pct = 1.0
     if max_sample_size:
         max_sample_size_pct = 1.0 * max_sample_size / np.shape(X_train)[0]
+    min_sample_size_pct = 1.0 * min_sample_size / np.shape(X_train)[0]
+    search_space = create_search_space(max_components, min_sample_size, min_sample_size_pct,
+                                       max_sample_size_as_pct=max_sample_size_pct,
+                                       extended_space=use_extended_search_space, optimizer_type=optimizer)
+    # if the optimizer is hyperopt, then we use the search space as is
+    # but there is some handling required for pysot
+    if optimizer in {'pysot', 'lipo'}:
+        logging.info(f"Preparing search space for {optimizer}.")
+        unif_search_space, log_transforms = search_space
+        # get the lower and upper bounds
+        lb, ub, transforms = [], [], []
+        for k, v in unif_search_space.items():
+            if k in log_transforms:
+                transforms.append(math.exp)
+            else:
+                transforms.append(None)
+            lb.append(v[0])
+            ub.append(v[1])
 
-    init_vals = [{'dp_alpha': 0.1,
-                  'sample_size_as_pct': max_sample_size_pct,
-                  'pct_from_original': 1.0,
-                  'prior_for_a_beta_A': 1.0,
-                  'prior_for_a_beta_B': 1.0,
-                  'prior_for_b_beta_A': 1.0,
-                  'prior_for_b_beta_B': 1.0
-                  }]
+    init_vals = [OrderedDict({'dp_alpha': 0.1,
+                              'sample_size_as_pct': max_sample_size_pct,
+                              'pct_from_original': 1.0,
+                              'prior_for_a_beta_A': 1.0,
+                              'prior_for_a_beta_B': 1.0,
+                              'prior_for_b_beta_A': 1.0,
+                              'prior_for_b_beta_B': 1.0
+                              })]
+    opt_var_key_order = init_vals[0].keys()
 
-    if additional_init_vals is not None:
-        init_vals += additional_init_vals
-    logging.info("%d init vals provided." % (len(init_vals)))
+    # prepare a wrapper object for the optimization objective function because many libraries
+    # require something like this
+    class WrapperOpt(OptimizationProblem):
+        def __init__(self):
+            # super.__init__()
+            self.dim = OPTVAR_DIMS
+            self.int_var = np.array([])
+            self.cont_var = np.arange(0, OPTVAR_DIMS)
+            self.lb, self.ub = np.array(lb), np.array(ub)
+            self.trials = []
 
-    trials = generate_trials_to_calculate(init_vals)
-    min_sample_size_pct = 1.0 * min_sample_size/np.shape(X_train)[0]
-    search_space = create_hyperopt_search_space(max_components, min_sample_size, min_sample_size_pct,
-                                                max_sample_size_as_pct=max_sample_size_pct,
-                                                extended_space=True)
-    best_params_train = fmin(minimization_objective,
-                             space=search_space,
-                             algo=tpe.suggest,
-                             max_evals=evals,
-                             trials=trials)
+        def eval(self, params):
+            # transform the params if needed - pysot only supports uniform spaces
+            tr_params = np.array([param if tr is None else tr(param) for param, tr in zip(params, transforms)])
+            result = minimization_objective(tr_params)
+            self.trials.append(result)
+            return result['loss']
+
+        def eval_d(self, **params):
+            """
+            This evaluates based on a dictionary of named variables. Required by LIPO.
+            :param params:
+            :return:
+            """
+            param_arr = np.array([params[k] for k in opt_var_key_order])
+            return self.eval(param_arr)
+
+    if optimizer == 'hyperopt':
+        if additional_init_vals is not None:
+            init_vals += additional_init_vals
+        logging.info("%d init vals provided." % (len(init_vals)))
+
+        trials = generate_trials_to_calculate(init_vals)
+
+        best_params_train = fmin(minimization_objective,
+                                 space=search_space,
+                                 algo=tpe.suggest,
+                                 max_evals=evals,
+                                 trials=trials)
+    elif optimizer == 'pysot':
+        init_vals = [np.array(list(init_vals[0].values()))]
+        if additional_init_vals is not None:
+            init_vals += additional_init_vals
+        init_vals = np.array(init_vals)
+        logging.info("%d init vals provided." % (len(init_vals)))
+        f = WrapperOpt()
+        # pySOT doesn't evaluate the init vals automatically, we need to explicitly provide these
+        logging.info(f"Obtaining response at init vals for pySOT, {len(init_vals)} evals needed.")
+        init_vals_y = [f.eval(x_init) for x_init in init_vals]
+
+        # set up a strategy and a controller
+        rbf = RBFInterpolant(dim=OPTVAR_DIMS, lb=f.lb, ub=f.ub, kernel=CubicKernel(), tail=LinearTail(7))
+        slhd = SymmetricLatinHypercube(dim=OPTVAR_DIMS, num_pts=2 * (OPTVAR_DIMS + 1))
+        controller = SerialController(f.eval)
+        pysot_strategy = 'srbf'
+        if optimizer_params is not None:
+            pysot_strategy = optimizer_params.get('strategy', 'srbf')
+        if pysot_strategy == 'srbf':
+            controller.strategy = SRBFStrategy(
+                max_evals=evals, opt_prob=f, exp_design=slhd, surrogate=rbf, asynchronous=True,
+                extra_points=init_vals, extra_vals=init_vals_y)
+        elif pysot_strategy == "dycors":
+            controller.strategy = DYCORSStrategy(
+                max_evals=evals, opt_prob=f, exp_design=slhd, surrogate=rbf, asynchronous=True,
+                extra_points=init_vals, extra_vals=init_vals_y)
+        else:
+            logging.error(f"Unknown pySOT strategy: {pysot_strategy}, aborting!!!")
+            return
+
+        # we also need to modify the controller fevals field for consistency
+        for extra_idx in range(len(init_vals)):
+            record = EvalRecord(params=(np.ravel(init_vals[extra_idx, :]),), status="completed")
+            record.value = init_vals_y[extra_idx]
+            record.feasible = True
+            controller.fevals.append(record)
+        logging.info(f"Updated controller records in pySOT - ready to run.")
+        pysot_result = controller.run()
+        logging.info(f"Finished pySOT run.")
+        temp_explored_space = np.array([(list(t.params[0]) + [t.value]) for t in controller.fevals])
+        visited_x, visited_y = temp_explored_space[:, :-1], temp_explored_space[:, -1]
+        best_params_train = visited_x[np.argmin(visited_y), :]
+
+    elif optimizer == "lipo":
+        # LIPO accepts the dictionary format of init vals
+        if additional_init_vals is not None:
+            init_vals += additional_init_vals
+        logging.info("%d init vals provided." % (len(init_vals)))
+        f = WrapperOpt()
+        # LIPO doesn't evaluate the init vals automatically, we need to explicitly provide these
+        logging.info(f"Obtaining response at init vals for LIPO, {len(init_vals)} evals needed.")
+        init_vals_y = [f.eval_d(**x_init) for x_init in init_vals]
+        search = GlobalOptimizer(
+            f.eval_d,
+            lower_bounds=dict([(var, bound) for var, bound in zip(opt_var_key_order, f.lb)]),
+            upper_bounds=dict([(var, bound) for var, bound in zip(opt_var_key_order, f.ub)]),
+            flexible_bounds=dict([(var, [False, False]) for var in opt_var_key_order]), # bounds should be fixed
+            evaluations=[(ix, iy) for ix, iy in zip(init_vals, init_vals_y)],
+            maximize=False  # we're minimizing loss
+        )
+        search.run(evals)
+        visited_x, visited_y = [], []
+        for es in search.evaluations:
+            visited_x.append([es[0][v] for v in opt_var_key_order])
+            visited_y.append(-es[1])  # remember to flip the sign, a LIPO oddity
+        visited_x, visited_y = np.array(visited_x), np.array(visited_y)
+        best_params_train = visited_x[np.argmin(visited_y), :]
+
+    trials_arr = None
+    if optimizer == 'hyperopt':
+        trials_arr = trials.results
+    if optimizer in {'pysot', 'lipo'}:
+        trials_arr = f.trials
 
     # write out the models
     suggested_model_path = "%s%sbest_mean_val_score_models" % (models_dir, os.sep)
@@ -730,30 +1016,42 @@ def run_experiment(dataset, model_obj, oracle,
     logging.info("Best params based on train: %s" % (best_params_train,))
     result_json["optimized"] = {}
     result_json["optimized"]["model_path"] = model_path if model_path else suggested_model_path
-    result_json["optimized"]["search_space"] = dict([(i, str(s)) for i, s in enumerate(search_space)])
+    if optimizer == 'hyperopt':
+        result_json["optimized"]["search_space"] = dict([(i, str(s)) for i, s in enumerate(search_space)])
+    if optimizer in {'pysot', 'lipo'}:
+        result_json["optimized"]["search_space"] = dict([(k, (f.lb[i], f.ub[i])) for i, k in enumerate(opt_var_key_order)])
     result_json["optimized"]["scale"] = scale
-    result_json["optimized"]["best_params_train"] = best_params_train
+    result_json["optimized"]["best_params_train"] = list(best_params_train)
     with open(trials_file, 'w') as f_trials:
-        f_trials.write(json.dumps(trials.results, indent=4))
+        f_trials.write(json.dumps(trials_arr, indent=4))
 
     # write the bootstrap score separately, for debugging purposes
     result_json["bootstrap"] = {}
-    result_json["bootstrap"]["mean_train_score"] = trials.results[0]["mean_train_score"]
-    result_json["bootstrap"]["mean_val_score"] = trials.results[0]["mean_val_score"]
-    result_json["bootstrap"]["mean_test_score"] = trials.results[0]["mean_test_score"]
-    result_json["bootstrap"]["loss"] = trials.results[0]["loss"]
-    result_json["bootstrap"]["params"] = init_vals[0]
+    result_json["bootstrap"]["mean_train_score"] = trials_arr[0]["mean_train_score"]
+    result_json["bootstrap"]["mean_val_score"] = trials_arr[0]["mean_val_score"]
+    result_json["bootstrap"]["mean_test_score"] = trials_arr[0]["mean_test_score"]
+    result_json["bootstrap"]["loss"] = trials_arr[0]["loss"]
+    if optimizer == 'hyperopt':
+        result_json["bootstrap"]["params"] = init_vals[0]
+    if optimizer in {'pysot', 'lipo'}:
+        result_json["bootstrap"]["params"] = dict([(i, j) for i, j in zip(opt_var_key_order, init_vals[0])])
 
     # find the best params on val
-    best_iter_idx_val = max(enumerate(trials.results), key=lambda t: t[1]["mean_val_score"])[0]
-    # tpe internally stores params in the following object as dicts, but each value is wrapped in a list
-    best_params_val = dict([(k, v[0])
-                                   for k, v in list(trials.trials[best_iter_idx_val]['misc']['vals'].items())])
-    result_json["optimized"]["best_params_val"] = best_params_val
+    best_iter_idx_val = max(enumerate(trials_arr), key=lambda t: t[1]["mean_val_score"])[0]
+    if optimizer == 'hyperopt':
+        # tpe internally stores params in the following object as dicts, but each value is wrapped in a list
+        best_params_val = dict([(k, v[0])
+                                for k, v in list(trials.trials[best_iter_idx_val]['misc']['vals'].items())])
+        best_iter_idx_train = min([(h['loss'], idx) for idx, h in enumerate(trials.results)])[1]
+        optimizer_scores = [i["mean_train_score"] for i in trials.results]
+    if optimizer in {'pysot', 'lipo'}:
+        best_params_val = visited_x[best_iter_idx_val, :]
+        best_iter_idx_train = int(np.argmin(visited_y))
+        optimizer_scores = visited_y
+
+    result_json["optimized"]["best_params_val"] = list(best_params_val)
     result_json["optimized"]["best_iter_idx_val"] = best_iter_idx_val
 
-    best_iter_idx_train = min([(h['loss'], idx) for idx, h in enumerate(trials.results)])[1]
-    optimizer_scores = [i["mean_train_score"] for i in trials.results]
     if compute_all_baselines is False:
         pct_changes_train, pct_changes_val = None, None
         best_train_improvement, best_val_improvement = None, None
@@ -761,13 +1059,12 @@ def run_experiment(dataset, model_obj, oracle,
         pct_changes_train = [100.0 * (s - baseline_train_score_based_on_train) / baseline_train_score_based_on_train
                                for s in optimizer_scores]
         pct_changes_val = [100.0 * (s["mean_val_score"] - baseline_val_score_based_on_train) /
-                                  baseline_val_score_based_on_train for s in trials.results]
+                           baseline_val_score_based_on_train for s in trials_arr]
         best_train_improvement = max(pct_changes_train)
         best_val_improvement = max(pct_changes_val)
 
     pct_changes_test = [100.0 * (s["mean_test_score"] - baseline_test_score_based_on_train_val) /
-                        baseline_test_score_based_on_train_val for s in trials.results]
-
+                        baseline_test_score_based_on_train_val for s in trials_arr]
     test_improvement_obj_fn_best_params_val = pct_changes_test[best_iter_idx_val]
     # test_improvement_best_params_val = 100.0 * (trials.results[best_iter_idx_val]["mean_test_score"]
     #                                             - baseline_test_score_based_on_train_val) / \
@@ -777,22 +1074,25 @@ def run_experiment(dataset, model_obj, oracle,
     logging.info("Baseline score: %0.04f" % (baseline_test_score_based_on_train_val))
 
     # sample_pct = 1.0 * sample_size/np.shape(X_train_val)[0]
-    result_json["optimized"]["total_trials"] = len(trials.results)
+    result_json["optimized"]["total_trials"] = len(trials_arr)
+    result_json["optimized"]["best_val_score"] = trials_arr[best_iter_idx_val]["mean_val_score"]
+    result_json["optimized"]["test_score_obj_fn_at_best_params_val"] = \
+        trials_arr[best_iter_idx_val]["mean_test_score"]
+    result_json["optimized"]["avg_train_complexity_best_params_val"] = \
+        trials_arr[best_iter_idx_val]["avg_complexity"]
+
     result_json["optimized"]["best_train_score"] = max(optimizer_scores)
-    result_json["optimized"]["best_val_score"] = trials.results[best_iter_idx_val]["mean_val_score"]
     result_json["optimized"]["improvement_best_train"] = best_train_improvement
     result_json["optimized"]["improvement_best_val"] = best_val_improvement
     result_json["optimized"]["improvement_test_obj_fn_best_params_val"] = \
         test_improvement_obj_fn_best_params_val
-    result_json["optimized"]["test_score_obj_fn_at_best_params_val"] = \
-        trials.results[best_iter_idx_val]["mean_test_score"]
+
     avg_complexity_for_best_model = model_obj.get_avg_complexity(best_model_so_far[0])
     result_json["optimized"]["avg_model_complexity_for_best_param_val"] = \
         str(avg_complexity_for_best_model)
     result_json["optimized"]["avg_model_complexity_for_best_param_val_pickled_hex"] = \
         pickle.dumps(avg_complexity_for_best_model).hex()
-    result_json["optimized"]["avg_train_complexity_best_params_val"] = \
-        trials.results[best_iter_idx_val]["avg_complexity"]
+
     result_json["optimized"]["best_iter_idx_train"] = best_iter_idx_train
     result_json["optimized"]["best_iter_idx_val"] = best_iter_idx_val
 
@@ -835,9 +1135,10 @@ def run_experiment(dataset, model_obj, oracle,
     # plot the train-val-test scores, to ensure no overfitting is happening
     fig = plt.figure()
     ax = fig.add_subplot(111)
-    ax.plot(list(range(len(trials.results))), [t["mean_train_score"] for t in trials.results], lw=0.5, label="train", c='green')
-    ax.plot(list(range(len(trials.results))), [t["mean_val_score"] for t in trials.results], lw=0.5, label="val", c='blue')
-    ax.plot(list(range(len(trials.results))), [t["mean_test_score"] for t in trials.results], lw=0.5, label="test", c='red')
+
+    ax.plot(list(range(len(trials_arr))), [t["mean_train_score"] for t in trials_arr], lw=0.5, label="train", c='green')
+    ax.plot(list(range(len(trials_arr))), [t["mean_val_score"] for t in trials_arr], lw=0.5, label="val", c='blue')
+    ax.plot(list(range(len(trials_arr))), [t["mean_test_score"] for t in trials_arr], lw=0.5, label="test", c='red')
     ax.set_xlabel('iterations')
     ax.set_ylabel('F1')
     # ax.set_title("test=%0.02f%%" %(test_improvement_best_params_val_report,))
@@ -872,6 +1173,7 @@ def run_experiment(dataset, model_obj, oracle,
     result_json["runtimes"]["total_runtime_in_secs"] = (end_time - start_time).total_seconds()
     with open(results_file, 'w') as f_res:
         f_res.write(json.dumps(result_json, indent=4, sort_keys=True))
+        f_res.flush()
     return test_improvement_obj_fn_best_params_val, best_params_val, new_persistence_obj
 
 
